# Copyright (c) Stas Shtin, 2010

# This file is part of IPyPBX.

# IPyPBX is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# IPyPBX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with IPyPBX.  If not, see <http://www.gnu.org/licenses/>.

from PyQt4 import QtCore, QtGui, QtSql


class BaseController(QtCore.QObject):
    """
    Base class for other controllers.

    Doesn't do anything useful on its own.
    """
    # TODO: possibly use a separate class for options and a meta-class.
    fields = ()
    view_list_fields = ()
    view_display_fields = ()
    view_display_fields_hidden = ()
    
    def __init__(self, model=None, view_list=None, view_display=None, parent=None, views=None):
        super(BaseController, self).__init__(parent=parent)

        self.views = views
        
        # Find out base name.
        classname = self.__class__.__name__
        self.basename = (
            classname[:-10] if classname.endswith('Controller')
            else classname)
        self.basename = self.basename[0].lower() + self.basename[1:]

        # Are we given an existing model?
        if model:
            self.model = model
        # Otherwise initialize a new model.
        else:
            self.model = QtSql.QSqlTableModel(parent)
            self.model.setTable('ipypbxweb_%s' % self.basename.lower())

            # Create model header from fields list.
            for i, field in enumerate(self.fields):
                self.model.setHeaderData(
                    i, QtCore.Qt.Horizontal, QtCore.QVariant(field))

            # Fetch model data.
            self.model.select()

        # Are we given an existing view list?
        if view_list:
            self.view_list = view_list
        # Otherwise get view list from the parent.            
        else:
            self.view_list = getattr(views, self.basename + 'ViewList')
            self.view_list.setModel(self.model)

            # Hide fields not meant for display.
            for i, field in enumerate(self.fields):
                if field not in self.view_list_fields:
                    self.view_list.hideColumn(i)

            # Stretch headers to fill all available width.
            self.view_list.setSelectionMode(QtGui.QTableView.SingleSelection)
            self.view_list.setSelectionBehavior(QtGui.QTableView.SelectRows)
            self.view_list.resizeColumnsToContents()
            self.view_list.resizeRowsToContents()
            self.view_list.horizontalHeader().setStretchLastSection(True)
        self.view_list.selectRow(0)

        # Are we given an existing view display?
        if view_display:
            self.view_display = view_display
        # Otherwise get view display from the parent.
        else:
            self.view_display = QtGui.QDataWidgetMapper(parent)
            self.view_display.setModel(self.model)

            # If view_display_fields is not send, display all fields except
            # the first one that is usually the ID.
            display_fields = self.getDisplayFields()
            for i, field in enumerate(self.fields):
                if field in display_fields:
                    field_widget = self.getFieldWidget(field)
                    self.view_display.addMapping(field_widget, i)

        # Select first row in the view list.
        self.view_display.toFirst()
        
        # Register signals for this controller.
        for data in self.getSignalsData():
            if len(data) == 3:
                sender, signal, receiver = data
                QtCore.QObject.connect(sender, QtCore.SIGNAL(signal), receiver)
            elif len(data) == 4:
                sender, signal, receiver, slot = data
                QtCore.QObject.connect(
                    sender, QtCore.SIGNAL(signal), receiver, QtCore.SLOT(slot))
                                       

    def getDisplayFields(self):
        """
        Return the list of fields to display.

        If it's not set explicitly, use all defined fields except the first
        which usually is the primary key.
        """
        # Do we have the fields to display explicitly set?
        if self.view_display_fields:
            fields = self.view_display_fields
        else:
            # Do we have hidden fields set?
            if self.view_display_fields_hidden:
                hidden = self.view_display_fields_hidden
            # If not, hide the first one.
            else:
                hidden = self.fields[0]

            # Filter away all hidden fields.
            fields = [field for field in self.fields if not field in hidden]
            
        return fields

    def getFieldWidget(self, field):
        """
        Return widget for given field name.
        """
        return getattr(
            self.views,
            self.basename + ''.join(word.capitalize()
                                    for word in field.split(' ')))

    def getSignalsData(self):
        """
        Default signals built from controller's base name.
        """
        # Default signals handle row selection, Add and Save buttons.
        return (
            (getattr(self.views, self.basename + 'Add'), 'clicked()', self.add),
            (self.view_list.selectionModel(),
             'currentRowChanged(QModelIndex,QModelIndex)',
             self.view_display, 'setCurrentModelIndex(QModelIndex)'),
            (getattr(self.views, self.basename + 'Save'), 'clicked()',
             self.save),
            )

    def add(self):
        """
        Add new object.
        """
        # Add a new row to list view.
        num_rows = self.model.rowCount()
        self.model.insertRows(num_rows, 1)
        self.view_list.selectRow(num_rows)

        # Disable adding more than one row.
        self.getFieldWidget('Add').setEnabled(False)

        # Focust to the first displayed field.
        self.getFieldWidget(self.getDisplayFields()[0]).setFocus()

        # Clear all displayed fields.
        for field in self.getDisplayFields():
            self.getFieldWidget(field).clear()
        
    def save(self):
        """
        Save to database.
        """
        # For now we just submit everything - QT seems to be able to handle it
        # on its own.
        self.view_display.submit()
        self.getFieldWidget('Add').setEnabled(True)
        


class ConnectionController(BaseController):
    """
    Connections controller.
    """
    fields = (
        'ID', 'Name', 'Local IP Address', 'Local Port',
        'Freeswitch IP Address', 'Freeswitch Port')
    view_list_fields = 'Name', 'Freeswitch IP Address', 'Freeswitch Port'
    
    def clone(self):
        """
        TODO: Clone an existing connection.

        This creates a new connection with bound data copied from another one.
        """

class SipProfileController(BaseController):
    """
    SIP Profile controller.
    """
    fields = (
        'ID', 'Connection ID', 'Name', 'External RTP IP', 'External SIP IP',
        'RTP IP', 'SIP IP', 'SIP Port', 'Accept Blind Registration',
        'Authenticate Calls', 'Is Active')
    view_list_fields = 'Name', 'SIP IP', 'SIP Port'
    view_display_fields_hidden = 'ID', 'Connection ID'


class DomainController(BaseController):
    """
    Domain controller.
    """
    fields = 'ID', 'SIP Profile ID', 'Host Name', 'Is Active'
    view_list_fields = 'SIP Profile ID', 'Host Name', 'Is Active'


class GatewayController(BaseController):
    """
    Gateway controller.
    """
    fields = (
        'ID', 'SIP Profile ID', 'Name', 'Username', 'Password', 'Realm',
        'From Domain', 'Expire In Seconds', 'Retry In Seconds',
        'Caller ID In From Field', 'Is Active')
    view_list_fields = 'SIP Profile ID', 'Name'


class EndpointController(BaseController):
    """
    Endpoint controller.
    """
    fields = 'ID', 'User ID', 'Password', 'Domain ID', 'Is Active'


class ExtensionController(BaseController):
    """
    Extension controller.
    """
    fields = (
        'ID', 'Destination Match', 'XML Dialplan', 'Domain ID', 'Endpoint ID',
        'Authenticate Calls', 'Is Active')
    view_list_fields = 'Destination Match', 'Domain ID', 'Endpoint ID'
        
